#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os, sys, time, random, copy, pprint
import termios, tty, fcntl, thread, curses, locale

ANSWER_YES  = 'o'
ANSWER_NO   = 'n'
ANSWER_QUIT = 'q'

DEFAULT_DATA_FILE = 'datas/boxworld!.txt'
DEFAULT_DATA_FILE = 'datas/GRIGoRusha.txt'
QUESTION_ENTER_FILE_DATA = "Entrez un fichier de données à charger (par défaut : %s) :"
data_file = DEFAULT_DATA_FILE

EXCEPTION_BAD_CHAR          = "Bad character in data file"
EXCEPTION_TWICE_THE_PLAYER  = "Data file: a player is declared twice in one level"
EXCEPTION_CASE_NOT_HANDLED  = "Case that should never happen not handled"
EXCEPTION_Y_OVERFLOW        = "Y overflow"
EXCEPTION_X_OVERFLOW        = "X overflow"
EXCEPTION_NO_PLAYER_FOUND   = "No player found"
EXCEPTION_NO_OBJECT_TO_MOVE = "No object to move"
EXCEPTION_MORE_THAN_TWO_OBJECTS_ON_A_CELL = "More than 2 objects on a cell"
EXCEPTION_INVALID_DRAWINGS_SET_NUMBER = "Invalid drawings set number"

# Different drawings:
glob_tab_chars = \
[
    # Ground / GroundStoneDest / Wall / Player   / Stone
    [ ' ',     '.',              '█',   '☺', '☻', '○','♦' ],
    [ ' ',     '.',              '#',   '@', '+', '$','*' ],
    [ ' ',     '.',              '░',   '☺', '☻', '○','◙' ]
]

class Ground(object):
    def __init__(self):
        super(Ground,self).__init__() # Parent call:

class GroundStoneDest(object):
    def __init__(self):
        super(GroundStoneDest,self).__init__() # Parent call:

class Wall(object):
    def __init__(self):
        super(Wall,self).__init__() # Parent call:

class Player(object):
    def __init__(self):
        super(Player,self).__init__() # Parent call:

class Stone(object):
    def __init__(self):
        super(Stone,self).__init__() # Parent call:

class Level:
    """ Class that handles a full level
        -------------------
        When created, the params :
        - tab_comments = lines of comments
        - raw_data = Full level read from a file
        -------------------
        For raw_data, only valid chars :
          ' ' = empty space
          '#' = wall
          '$' = a stone
          '.' = where to put a stone ('stone place')
          '*' = a stone well placed
          '@' = player starting pos.
          '+' = player starting pos. on a 'stone place'
    """
    def __init__(self, tab_comments, raw_data, stdscr):
        self.tab_comments = copy.copy(tab_comments)
        self.player_pos = [-1, -1]
        self.player_moves = []
        self.data = []
        self.stones_total = 0
        self.stones_ok    = 0
        self.stones_dest  = 0
        # raw_data = Full level read from a file
        y = 0
        for line in raw_data:
            self.data.append([])
            for char in line:
                if char==' ': # ' ' = empty space
                    self.data[y].append([Ground()])
                elif char=='#': # '#' = wall
                    self.data[y].append([Wall()])
                elif char=='$': # '$' = a stone
                    self.data[y].append([Ground(),Stone()])
                    self.stones_total = self.stones_total + 1
                elif char=='.': # '.' = where to put a stone ('stone dest')
                    self.data[y].append([GroundStoneDest()])
                    self.stones_dest = self.stones_dest + 1
                elif char=='*': # '*' = a stone well placed
                    self.data[y].append([GroundStoneDest(),Stone()])
                    self.stones_total = self.stones_total + 1
                    self.stones_ok    = self.stones_ok + 1
                    self.stones_dest  = self.stones_dest + 1
                elif char=='@': # '@' = player starting pos.
                    if self.player_pos != [-1,-1]:
                        raise Exception(EXCEPTION_TWICE_THE_PLAYER)
                    self.player_pos = [y, len(self.data[y]) ]
                    self.data[y].append([Ground(),Player()])
                elif char=='+': # '+' = player starting pos. on a 'stone dest'
                    if self.player_pos != [-1,-1]:
                        raise Exception(EXCEPTION_TWICE_THE_PLAYER)
                    self.player_pos = [y, len(self.data[y]) ]
                    self.data[y].append([GroundStoneDest(),Player()])
                    self.stones_dest  = self.stones_dest + 1
                else:
                   # Should never happen but... we never know:
                   raise Exception(EXCEPTION_BAD_CHAR)
            y = y + 1

        if self.player_pos == [-1,-1]:
            raise Exception(EXCEPTION_NO_PLAYER_FOUND)

        # We got here = no error
        self.apply_drawings_set(0)
        self.apply_drawings_set(1)

    def apply_drawings_set(self, no_set):
        global glob_tab_chars
        if (no_set<-1) or (no_set>=len(glob_tab_chars)):
            raise Exception(EXCEPTION_INVALID_DRAWINGS_SET_NUMBER)
        if no_set>=0:
            no = no_set
        else:
            no = 0
        for y, tab_objs in enumerate(self.data):
            for x, objs in enumerate(tab_objs):
                for obj in objs:
                    # Calculate links to other walls:
                    if isinstance(obj, Ground):
                        obj.char = glob_tab_chars[no][0]
                    elif isinstance(obj, GroundStoneDest):
                        obj.char = glob_tab_chars[no][1]
                    elif isinstance(obj, Player):
                        obj.char           = glob_tab_chars[no][3]
                        obj.char_highlight = glob_tab_chars[no][4]
                    elif isinstance(obj, Stone):
                        obj.char           = glob_tab_chars[no][5]
                        obj.char_highlight = glob_tab_chars[no][6]
                    elif isinstance(obj, Wall):
                        if no_set>=0:
                            obj.char = glob_tab_chars[no][2]
                        else:
                            links = 0
                            if y>0:
                                if x<len(self.data[y-1]):
                                    if isinstance(self.data[y-1][x][0], Wall):
                                        links = links | 1
                            if y<(len(self.data)-1):
                                if x<len(self.data[y+1]):
                                    if isinstance(self.data[y+1][x][0], Wall):
                                        links = links | 4
                            if x>0:
                                if isinstance(self.data[y][x-1][0], Wall):
                                    links = links | 8
                            if x<(len(tab_objs)-1):
                                if isinstance(self.data[y][x+1][0], Wall):
                                    links = links | 2

                            if links==1:    uni = 9553 # <║>
                            elif links==2:  uni = 9552 # <═>
                            elif links==3:  uni = 9562 # <╚>
                            elif links==4:  uni = 9553 # <║>
                            elif links==5:  uni = 9553 # <║>
                            elif links==6:  uni = 9556 # <╔>
                            elif links==7:  uni = 9568 # <╠>
                            elif links==8:  uni = 9552 # <═>
                            elif links==9:  uni = 9565 # <╝>
                            elif links==10: uni = 9552 # <═>
                            elif links==11: uni = 9577 # <╩>
                            elif links==12: uni = 9559 # <╗>
                            elif links==13: uni = 9571 # <╣>
                            elif links==14: uni = 9574 # <╦>
                            else:           uni = 9580 # <╬>
                            obj.char=unichr(uni).encode('utf-8')

    def draw(self, stdscr, start_y, start_x):
        for y, tab_objs in enumerate(self.data):
            for x, objs in enumerate(tab_objs):
                if len(objs)==1:
                    stdscr.addstr(start_y+y, start_x+x, objs[0].char)
                elif len(objs)==2:
                    if isinstance(objs[0], GroundStoneDest):
                        # Draw on stone destination tile
                        stdscr.addstr(start_y+y, start_x+x, objs[1].char_highlight)
                    elif isinstance(objs[0], Ground):
                        # Draw on empty tile
                        stdscr.addstr(start_y+y, start_x+x, objs[1].char)
                    else:
                        # Should never happen but... we never know:
                        raise Exception(EXCEPTION_CASE_NOT_HANDLED)

    def register_player_move(self, add_y, add_x, stone_pushed):
        self.player_moves.append([add_y, add_x, stone_pushed])

    def move(self, y, x, add_y, add_x, call_depth = 1):
        if (y<0) or (y>=len(self.data)):
            raise Exception(EXCEPTION_Y_OVERFLOW)
        if (x<0) or (x>=len(self.data[y])):
            raise Exception(EXCEPTION_X_OVERFLOW)
        if ((y+add_y)<0) or ((y+add_y)>=len(self.data)):
            return False
        if ((x+add_x)<0) or ((x+add_x)>=len(self.data[y])):
            return False
        # No overflow = test possible
        if len(self.data[y][x])<2:
            # Should never happen but... we never know:
            raise Exception(EXCEPTION_NO_OBJECT_TO_MOVE)
        tab_src=self.data[y][x]
        tab_dst=self.data[y+add_y][x+add_x]
        if len(tab_dst)==0:
            return True
        elif len(tab_dst)==1:
            # One object = Wall, Ground or GroundStoneDest
            if isinstance(tab_dst[0], GroundStoneDest) or \
                isinstance(tab_dst[0], Ground):
                # Before moving, see if we add well placed stone:
                if isinstance(tab_src[1], Stone):
                    if isinstance(tab_src[0], GroundStoneDest):
                        self.stones_ok = self.stones_ok - 1
                    if isinstance(tab_dst[0], GroundStoneDest):
                        self.stones_ok = self.stones_ok + 1
                # Move the object from (y,x) to (y+add_y,x+add_x)
                if isinstance(tab_src[1], Player):
                    self.register_player_move(add_y, add_x, False)
                self.data[y+add_y][x+add_x].append(self.data[y][x][1])
                self.data[y][x]=[self.data[y][x][0]]
                return True
            elif isinstance(tab_dst[0], Wall):
                return False
            # Should never happen but... we never know:
            raise Exception(EXCEPTION_CASE_NOT_HANDLED)
        elif len(tab_dst)==2:
            if call_depth>1: # Already testing if a stone cane move
                return False
            if self.move(y+add_y, x+add_x, add_y, add_x, call_depth+1):
                # The stone has moved => move here, too
                # Move the object from (y,x) to (y+add_y,x+add_x)
                if isinstance(tab_src[1], Player):
                    self.register_player_move(add_y, add_x, True)
                self.data[y+add_y][x+add_x].append(self.data[y][x][1])
                self.data[y][x]=[self.data[y][x][0]]
                return True
            return False
        # We should never reach here but... we never know:
        raise Exception(EXCEPTION_MORE_THAN_TWO_OBJECTS_ON_A_CELL)

    def solved(self):
        if self.stones_total < self.stones_dest:
            return self.stones_ok == self.stones_total
        return self.stones_ok == self.stones_dest

class LevelsReader:

    """ Class that loads a file filled with levels
        -------------------
        Files must be like this:
        -------------------
        Whatever comment then (as many lines as you
        want), then for the level itself,
        lines with only valid chars :
          ' ' = empty space
          '#' = wall
          '$' = a stone
          '.' = where to put a stone ('stone place')
          '*' = a stone well placed
          '@' = player starting pos.
          '+' = player starting pos. on a 'stone place'
    """

    def __init__(self, file_name,stdscr):
        # Array of Level objects:
        self.tab = []
        self.tab_original = []
        self.CleanTemporaryLevel()
        f = open(file_name, 'r')
        look_for_start = True
        for line in f:
            line = line.rstrip()
            if look_for_start:
                if len(line)>0:
                    if self.OnlyMazeChars(line):
                        look_for_start = False
                        self.CleanTemporaryLevel()
                        self.AddLine(line)
                    else:
                        self.AddComment(line)
            else:
                if len(line)>0:
                    if self.OnlyMazeChars(line):
                        self.AddLine(line)
                    else:
                        look_for_start = True
                        self.AddTemporaryLevel(stdscr)
                        self.CleanTemporaryLevel()
                        self.AddComment(line)
                else:
                    self.AddTemporaryLevel(stdscr)
                    self.CleanTemporaryLevel()
        # Finished read, but don't forget last level:
        if len(self.tmp_level)>0:
            self.AddTemporaryLevel(stdscr)

    def OnlyMazeChars(self, line):
        for char in line:
            if ord(char)==10:
                continue
            if not (char in ['#','$','.','*','@','+',' ']) :
                return False
        return True

    def AddComment(self, line):
        self.tmp_comments.append(line)

    def AddTemporaryLevel(self,stdscr):
        if len(self.tmp_level)>0:
            self.tab.append(
                Level(self.tmp_comments, self.tmp_level,stdscr)
            )
            self.tab_original.append(
                Level(self.tmp_comments, self.tmp_level,stdscr)
            )

    def CleanTemporaryLevel(self):
        self.tmp_comments = []
        self.tmp_level = []

    def AddLine(self, line):
        self.tmp_level.append(line)


def curses_main(stdscr, code):
    """ Main """
    global data_file

    curses.savetty()
    try:
        # Hide the cursor:
        curses.curs_set(0)
        levels = LevelsReader(data_file,stdscr)

        no_current_level = 30
        current_drawings_set = 2
        offset_x = 10
        offset_y = 18

        finished = False
        while not finished:
            stdscr.clear()
            level = levels.tab[no_current_level]
            stdscr.addstr( 1, 1, "'az' - change display")
            stdscr.addstr( 2, 1, "'-p' - prev level")
            stdscr.addstr( 3, 1, "'+n' - next level")
            stdscr.addstr( 4, 1, "'r'  - reset current level")
            stdscr.addstr( 5, 1, "'h'  - left")
            stdscr.addstr( 6, 1, "'l'  - right")
            stdscr.addstr( 7, 1, "'k'  - up")
            stdscr.addstr( 8, 1, "'j'  - down")
            stdscr.addstr( 9, 1, "'u'  - undo")
            stdscr.addstr(10, 1, "'q'  - quit")
            level.apply_drawings_set(current_drawings_set)
            level.draw(stdscr, offset_y, offset_x)
            stdscr.refresh()

            (xmin,ymin) = stdscr.getbegyx()
            (xmax,ymax) = stdscr.getmaxyx()

            stdscr.refresh()

            # halfdelay() = "set getch() to non blocking",
            # halfdelay(1): wait delay = 1/10 second:
            curses.halfdelay(1)
            while True:
                try_to_move = False
                try:
                    c=stdscr.getch()
                except:
                    c=-1
                if c<0:
                    pass
                elif 0<c<256:
                    if chr(c) in 'kK': try_to_move=(-1, 0)
                    elif chr(c) in 'jJ': try_to_move=( 1, 0)
                    elif chr(c) in 'hH': try_to_move=( 0,-1)
                    elif chr(c) in 'lL': try_to_move=( 0, 1)
                    elif chr(c) in 'aA':
                        current_drawings_set = current_drawings_set + 1
                        if current_drawings_set == len(glob_tab_chars):
                            current_drawings_set = -1
                        break
                    elif chr(c) in '-Pp':
                        no_current_level = no_current_level - 1
                        break
                    elif chr(c) in '+Nn':
                        no_current_level = no_current_level + 1
                        break
                    elif chr(c) in 'Qq':
                        finished = True
                        break
                    else:
                        stdscr.addstr(0,50, curses.keyname(c)+ ' pressed')
                        stdscr.refresh()
                elif c==curses.KEY_UP:    try_to_move=(-1, 0)
                elif c==curses.KEY_DOWN:  try_to_move=( 1, 0)
                elif c==curses.KEY_LEFT:  try_to_move=( 0,-1)
                elif c==curses.KEY_RIGHT: try_to_move=( 0, 1)
                elif c==curses.KEY_HOME:
                    pass
                elif c==curses.KEY_END:
                    pass
                else:
                    pass
                if try_to_move != False:
                    (y, x) = level.player_pos
                    (add_y, add_x) = try_to_move
                    if level.move(y,x,add_y,add_x):
                        level.player_pos = (y+add_y, x+add_x)
                        level.draw(stdscr, offset_y, offset_x)
                        stdscr.refresh()
                        if level.solved():
                            no_current_level = no_current_level + 1
                            break
    finally:
        curses.resetty()

def main():
    locale.setlocale(locale.LC_ALL, '')
    code = locale.getpreferredencoding()
    # Launch main app through curses wrapper
    # (=> if there's an Exception, quits properly):
    curses.wrapper(curses_main, code)

# If the script is executed, launch main function:
if __name__ == '__main__':
    main()

